# Imports
import logging
import networkx as nx

from collections import OrderedDict

from kadmos.graph import RepositoryConnectivityGraph, FundamentalProblemGraph
from kadmos.utilities.general import get_mdao_setup

# Settings for logging
logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)

# List of MDAO definitions that can be wrapped around the problem
mdao_definitions = ['unconverged-MDA-GS',  # 0
                    'unconverged-MDA-J',   # 1
                    'converged-MDA-GS',    # 2
                    'converged-MDA-J',     # 3
                    'unconverged-DOE-GS',  # 4
                    'unconverged-DOE-J',   # 5
                    'converged-DOE-GS',    # 6
                    'converged-DOE-J',     # 7
                    'unconverged-OPT-GS',  # 8
                    'unconverged-OPT-J',   # 9
                    'MDF-GS',              # 10
                    'MDF-J',               # 11
                    'IDF']                 # 12

# Settings for scripting
mdao_definitions_loop_all = False      # Option for looping through all MDAO definitions
mdao_definition_id = 10                 # Option for selecting a MDAO definition (in case mdao_definitions_loop_all=False)

# Settings for saving
pdf_dir = 'variable_complexity_problem/(X)DSM'
cmdows_dir = 'variable_complexity_problem/CMDOWS'
kdms_dir = 'variable_complexity_problem/KDMS'
vistoms_dir = 'variable_complexity_problem/VISTOMS'

while True:
    # Settings for variable complexity problem
    n_disciplines = 5
    coupling_strength = 0.1
    feedback_required = True
    problem_formulation = 'random'  # ['manual', 'random']

    print 'Scripting RCG...'

    # A new repository connectivity graph is defined to describe the general problem
    rcg = RepositoryConnectivityGraph(name='Variable complexity problem graph')
    # A description is added to the graph
    rcg.graph['description'] = 'Repository graph of the variable complexity problem'

    # Create a random mathematical problem
    math_problem = rcg.create_mathematical_problem(n_disciplines, coupling_strength)

    # Check if feedback couplings exist
    if nx.is_directed_acyclic_graph(rcg) and feedback_required and problem_formulation == 'random':
        print 'No feedback loops exist in formulated problem. New problem is generated...'
        continue
    break

# Create a DSM visualization of the RCG
rcg.create_dsm(file_name='RCG', function_order=math_problem['function_order'], include_system_vars=True, destination_folder=pdf_dir)
# Create a VISTOMS visualization of the RCG
rcg.vistoms_create(vistoms_dir, function_order=math_problem['function_order'], compress=False)

# Save the RCG as kdms
rcg.save('RCG', destination_folder=kdms_dir)
# Save the RCG as cmdows (and do an integrity check)
rcg.save('RCG', file_type='cmdows', destination_folder=cmdows_dir,
         description='RCG CMDOWS file of the variable complexity problem',
         creator='Anne-Liza Bruggeman',
         version='0.1',
         pretty_print=True,
         integrity=True)

print 'Scripting initial FPG...'

# An initial fundamental problem graph is created based on the rcg
fpg_initial = FundamentalProblemGraph(rcg)

# On to the wrapping of the MDAO architectures
# Get iterator (all or single one)
if not mdao_definitions_loop_all:
    mdao_definitions = [mdao_definitions[mdao_definition_id]]

for mdao_definition in mdao_definitions:

    print 'Scripting ' + str(mdao_definition) + '...'

    # Determine the three main settings: architecture, convergence type and unconverged coupling setting
    mdao_architecture, convergence_type, allow_unconverged_couplings = get_mdao_setup(mdao_definition)

    # Reset FPG
    fpg = fpg_initial.cleancopy()
    fpg.graph['name'] = rcg.graph['name'] + ' - ' + mdao_definition + ' - FPG'
    fpg.graph['description'] = 'Fundamental problem graph for solving the variable complexity problem using the strategy: ' \
                               + mdao_definition + '.'

    # Get possible function order
    function_order = fpg.get_possible_function_order(method='brute-force')

    # Define settings of the problem formulation
    fpg.graph['problem_formulation'] = dict()
    fpg.graph['problem_formulation']['function_order'] = function_order
    fpg.graph['problem_formulation']['mdao_architecture'] = mdao_architecture
    fpg.graph['problem_formulation']['convergence_type'] = convergence_type
    fpg.graph['problem_formulation']['allow_unconverged_couplings'] = allow_unconverged_couplings
    if mdao_architecture in ['unconverged-DOE', 'converged-DOE']:
        fpg.graph['problem_formulation']['doe_settings'] = dict()
        fpg.graph['problem_formulation']['doe_settings']['doe_method'] = 'Custom design table'
        if fpg.graph['problem_formulation']['doe_settings']['doe_method'] in ['Latin hypercube design',
                                                                              'Monte Carlo design']:
            fpg.graph['problem_formulation']['doe_settings']['doe_seed'] = 6
            fpg.graph['problem_formulation']['doe_settings']['doe_runs'] = 5
        elif fpg.graph['problem_formulation']['doe_settings']['doe_method'] in ['Full factorial design']:
            fpg.graph['problem_formulation']['doe_settings']['doe_runs'] = 5

    # Depending on the architecture, different additional node attributes have to be specified. This is automated here
    # to allow direct execution of all different options.
    if mdao_architecture in ['IDF', 'MDF', 'unconverged-OPT']:
        fpg.mark_as_objective('/data_schema/objective/f')
        for constraint in range(math_problem['n_global_constraints']):
            fpg.mark_as_constraint('/data_schema/constraints/g{0}'.format(constraint+1), '<=', 0.0)
        for global_var in range(math_problem['n_global_var']):
            fpg.mark_as_design_variable('/data_schema/global_design_variables/x0{0}'.format(global_var+1),
                                        lower_bound=-5, upper_bound=5, nominal_value=0)
        for discipline in range(math_problem['n_disciplines']):
            for local_var in range(math_problem['n_local_var'][discipline]):
                fpg.mark_as_design_variable('/data_schema/local_design_variables/x{0}{1}'.format(discipline+1, local_var+1),
                                            lower_bound=-5, upper_bound=5, nominal_value=0)
    elif mdao_architecture in ['unconverged-MDA', 'converged-MDA']:
        fpg.mark_as_qois(['/data_schema/objective/f'])
        for discipline in range(math_problem['n_disciplines']):
            for coupling_var in range(math_problem['n_coupling_var'][discipline]):
                fpg.mark_as_qois(['/data_schema/coupling_variables/y{0}{1}'.format(discipline+1, coupling_var+1)])
        for constraint in range(math_problem['n_global_constraints']):
            fpg.mark_as_qois(['/data_schema/constraints/g{0}'.format(constraint+1)])
    elif mdao_architecture in ['unconverged-DOE', 'converged-DOE']:
        fpg.mark_as_qois(['/data_schema/objective/f'])
        for constraint in range(math_problem['n_global_constraints']):
            fpg.mark_as_qois(['/data_schema/constraints/g{0}'.format(constraint+1)])
        for global_var in range(math_problem['n_global_var']):
            fpg.mark_as_design_variable('/data_schema/global_design_variables/x0{0}'.format(global_var+1),
                                        samples=[1.0, 2.0, 3.0, 4.0], lower_bound=-5, upper_bound=5, nominal_value=0)
        for discipline in range(math_problem['n_disciplines']):
            for local_var in range(math_problem['n_local_var'][discipline]):
                fpg.mark_as_design_variable('/data_schema/local_design_variables/x{0}{1}'.format(discipline+1, local_var+1),
                                            samples=[1.0, 2.0, 3.0, 4.0], lower_bound=-5, upper_bound=5, nominal_value=0)

    # Search for problem roles
    fpg.add_function_problem_roles()

    # Create a DSM visualization of the FPG
    fpg.create_dsm(file_name='FPG_'+mdao_definition, function_order=function_order, include_system_vars=True,
                   destination_folder=pdf_dir)
    # Create a VISTOMS visualization of the FPG (and add it to the existing directory)
    fpg.vistoms_add(vistoms_dir, function_order=function_order)

    # Save the FPG as kdms
    fpg.save('FPG_'+mdao_definition, destination_folder=kdms_dir)
    # Save the FPG as cmdows (and do an integrity check)
    fpg.save('FPG_'+mdao_definition, file_type='cmdows', destination_folder=cmdows_dir,
             description='FPG CMDOWS file of the variable complexity problem',
             creator='Anne-Liza Bruggeman',
             version='0.1',
             pretty_print=True,
             integrity=True)

    # Get Mdao graphs
    mpg = fpg.get_mpg(name='mpg variable complexity problem')
    mdg = fpg.get_mdg(name='mdg variable complexity problem')
    mdg.graph['name'] = rcg.graph['name'] + ' - ' + mdao_definition + ' - Mdao'
    mdg.graph['description'] = 'Solution strategy to solve the variable complexity problem using the strategy: ' \
                               + str(mdao_architecture) + ('_' + str(convergence_type) if convergence_type else '') + '.'

    # Copy mathematical functions to additional output edges (for MDF and IDF additional output edges are created)
    # TODO: This should be automated in the KADMOS method get_mdg()
    if mdao_architecture in ['IDF', 'MDF', 'unconverged-OPT']:
        mdg.edge['F']['/data_schema/architectureNodes/finalOutputVariables/data_schemaCopy/objective/f']['equations'] = mdg.edge['F']['/data_schema/objective/f']['equations']
        for constraint in range(math_problem['n_global_constraints']):
            mdg.edge['G{0}'.format(constraint+1)]['/data_schema/architectureNodes/finalOutputVariables/data_schemaCopy/constraints/g{0}'.format(constraint+1)]['equations'] =\
                mdg.edge['G{0}'.format(constraint+1)]['/data_schema/constraints/g{0}'.format(constraint+1)]['equations']
    for discipline in range(math_problem['n_disciplines']):
        for coupling_var in range(math_problem['n_coupling_var'][discipline]):
            if mdg.has_node('/data_schema/architectureNodes/finalCouplingVariables/data_schemaCopy/coupling_variables/y{0}{1}'.format(discipline+1, coupling_var+1)):
                mdg.edge['D{0}'.format(discipline+1)]['/data_schema/architectureNodes/finalCouplingVariables/data_schemaCopy/coupling_variables/y{0}{1}'.format(discipline+1, coupling_var+1)]['equations'] = \
                    mdg.edge['D{0}'.format(discipline + 1)]['/data_schema/coupling_variables/y{0}{1}'.format(discipline+1, coupling_var+1)]['equations']

    # Create a DSM visualization of the Mdao
    mdg.create_dsm(file_name='Mdao_'+mdao_definition, include_system_vars=True, destination_folder=pdf_dir, mpg=mpg)
    # Create a VISTOMS visualization of the Mdao (and add it to the existing directory)
    mdg.vistoms_add(vistoms_dir, mpg=mpg)

    # Save the Mdao as kdms
    mdg.save('Mdao_'+mdao_definition, destination_folder=kdms_dir, mpg=mpg)
    # Save the Mdao as cmdows (and do an integrity check)

    mdg.add_equation_labels(mdg.get_function_nodes(), labeling_method='node_id')

    mdg.save('Mdao_'+mdao_definition, file_type='cmdows', destination_folder=cmdows_dir,
             mpg=mpg,
             description='Mdao CMDOWS file of the variable complexity problem',
             creator='Anne-Liza Bruggeman',
             version='0.1',
             pretty_print=True,
             integrity=True
             )

print 'Done!'